/* System-level routines.
 *	- exception entry points
 *	- kernel image entry point
 *	- space for a temporary stack
 *	- a storage place for system registers
 */

        .text

/*
 * Disable "intelligent" assembler - it will be more clear for us.
 */
        .set    noat
        .set    noreorder
        .set    nomacro


/*
 * Include definitions of register names and other useful constants.
 */
#include "regname.h"
#include "offset.h"

/*
 * Offsets of the system registers as they are stored at the safe place.
 */
#define OFFSET_SAFE_EPC         0x00
#define OFFSET_SAFE_CAUSE       0x04
#define OFFSET_SAFE_BADVADDR    0x08
#define OFFSET_SAFE_STATUS      0x0C


/*
 * Define macros for MSIM debugging instructions.
 *
 * drv		dumps the contents of all general registers
 * dhlt		stops the simulator
 */
        .macro	drv
        .insn
        .word	0x37
        .endm	drv

        .macro	dhlt
        .insn
        .word	0x28
        .endm	dhlt



/*
 * MACRO
 * Store all general registers into the stack frame. Skipping k0, k1 and sp -
 * we don't need them.
 */
        .macro  STORE_GENERAL_REGS base

        sw      $0,  OFFSET_ZERO(\base)	/* yes, you are right... */

        sw      $at, OFFSET_AT(\base)

        sw      $v0, OFFSET_V0(\base)
        sw      $v1, OFFSET_V1(\base)

        sw      $a0, OFFSET_A0(\base)
        sw      $a1, OFFSET_A1(\base)
        sw      $a2, OFFSET_A2(\base)
        sw      $a3, OFFSET_A3(\base)

        sw      $t0, OFFSET_T0(\base)
        sw      $t1, OFFSET_T1(\base)
        sw      $t2, OFFSET_T2(\base)
        sw      $t3, OFFSET_T3(\base)
        sw      $t4, OFFSET_T4(\base)
        sw      $t5, OFFSET_T5(\base)
        sw      $t6, OFFSET_T6(\base)
        sw      $t7, OFFSET_T7(\base)

        sw	    $s0, OFFSET_S0(\base)
        sw	    $s1, OFFSET_S1(\base)
        sw	    $s2, OFFSET_S2(\base)
        sw	    $s3, OFFSET_S3(\base)
        sw	    $s4, OFFSET_S4(\base)
        sw	    $s5, OFFSET_S5(\base)
        sw	    $s6, OFFSET_S6(\base)
        sw	    $s7, OFFSET_S7(\base)

        sw	    $t8, OFFSET_T8(\base)
        sw	    $t9, OFFSET_T9(\base)

        sw	    $gp, OFFSET_GP(\base)
        sw	    $fp, OFFSET_FP(\base)
        sw	    $ra, OFFSET_RA(\base)

        .endm   STORE_GENERAL_REGS

/*
 * MACRO
 * Restore all general registers into the stack frame. Skip k0, k1 and sp.
 */
        .macro  RESTORE_GENERAL_REGS base

        lw      $ra, OFFSET_RA(\base)
        lw      $fp, OFFSET_FP(\base)
        lw      $gp, OFFSET_GP(\base)

        lw      $t9, OFFSET_T9(\base)
        lw      $t8, OFFSET_T8(\base)

        lw      $s7, OFFSET_S7(\base)
        lw      $s6, OFFSET_S6(\base)
        lw      $s5, OFFSET_S5(\base)
        lw      $s4, OFFSET_S4(\base)
        lw      $s3, OFFSET_S3(\base)
        lw      $s2, OFFSET_S2(\base)
        lw      $s1, OFFSET_S1(\base)
        lw      $s0, OFFSET_S0(\base)

        lw      $t7, OFFSET_T7(\base)
        lw      $t6, OFFSET_T6(\base)
        lw      $t5, OFFSET_T5(\base)
        lw      $t4, OFFSET_T4(\base)
        lw      $t3, OFFSET_T3(\base)
        lw      $t2, OFFSET_T2(\base)
        lw      $t1, OFFSET_T1(\base)
        lw      $t0, OFFSET_T0(\base)

        lw      $a3, OFFSET_A3(\base)
        lw      $a2, OFFSET_A2(\base)
        lw      $a1, OFFSET_A1(\base)
        lw      $a0, OFFSET_A0(\base)

        lw      $v1, OFFSET_V1(\base)
        lw      $v0, OFFSET_V0(\base)

        lw      $at, OFFSET_AT(\base)

        lw      $0,  OFFSET_ZERO(\base)	/* once again... */

        .endm   RESTORE_GENERAL_REGS



/*
 *
 * TLB Refill exception entry point
 *
 * Located at fixed address 0x00000000. The PC register is or-ed with
 * 0x80000000 to avoid mapping through TLB. Right now it only displays the
 * contents of registers and halts the machine.
 *
 */

        .org    0x0
        .ent    exception_tlbrefill

exception_tlbrefill:
        j       handle_tlbrefill
        nop

        .end    exception_tlbrefill



/*
 *
 * Cache Error exception entry point
 *
 * Located at fixed address 0x00000100. The PC register is or-ed with
 * 0x80000000 to avoid mapping through TLB. This exception cannot occur
 * in the simulated environment. Display registers and halt the machine.
 */

        .org    0x100
        .ent    exception_cacheerror

exception_cacheerror:
        drv
        dhlt

        .end    exception_cacheerror



/*
 *
 * General exception entry point
 *
 * Located at fixed address 0x00000180. The PC register is or-ed with
 * 0x80000000 to avoid mapping through TLB.
 */
        .org    0x180
        .ent    exception_general

exception_general:
        j	      handle_general_exception
        nop

        .end    exception_general



/*
 * A temporary stack is used in the first kernel initialize and by
 * the TLB routines. Located at fixed address 0x00000200 and 0x100 bytes
 * long.
 *
 * A top of the stack is also used as a safe place for storing some registers
 * from CP0.
 */
#define TEMP_STACK_TOP      0x80000200
#define TEMP_STACK_SIZE     0x100
#define SAFE_PLACE          TEMP_STACK_TOP

        .org    0x200
        .space  TEMP_STACK_SIZE



/*
 *
 * Kernel entry point
 *
 * Located at fixed address 0x0000300. The control is redirected here by
 * the loader. Our goal is to init basic environment such as stack and
 * jump to a C code.
 */
        .org    0x300
        .ent    __start
__start:

        /*
         * Set the temporary stack to our free place. Set the global
         * pointer to point to the kernel space (0x80000000).
         */
         lui    $sp, (TEMP_STACK_TOP + TEMP_STACK_SIZE) >> 16
         or     $sp, (TEMP_STACK_TOP + TEMP_STACK_SIZE) & 0xffff
         lui    $gp, 0x8000
         or     $gp, $0, $0

        /*
         * Jump to kernel C code.
         */
         jal	start
         nop

         .end __start



/*
 *
 * General exception handler
 */
        .ent    handle_general_exception

handle_general_exception:
        /*
         * Save EPC, Status, BadVAddr and Cause registers to a safe place.
         * We will need them later.
         */
        lui	$k1, (SAFE_PLACE) >> 16
        or	$k1, (SAFE_PLACE) & 0xffff

        mfc0	$k0, $epc
        sw      $k0, OFFSET_SAFE_EPC($k1)
        mfc0	$k0, $cause
        sw      $k0, OFFSET_SAFE_CAUSE($k1)
        mfc0    $k0, $badvaddr
        sw      $k0, OFFSET_SAFE_BADVADDR($k1)
        mfc0    $k0, $status
        sw      $k0, OFFSET_SAFE_STATUS($k1)

        /*
         * Leave exception level and disable interrupts. This is because we
         * need to store general registers on the stack frame which could
         * cause a TLB exception. A recursive exception call would overwrite
         * the return address and the kernel registers (k0, k1).
         */
        lui     $k1, 0xffff
        ori     $k1, $k1, 0xffc0 /* disable interrupts & leave exception level */
        and     $k0, $k1
        mtc0	$k0, $status

        /*
         * Allocate a frame on the stack and store all general registers
         * on the stack. Registers k0 and k1 are kernel scratch registers
         * which are not supposed to be stored and restored.
         */
        addiu   $sp, -EX_STACK_FRAME

        STORE_GENERAL_REGS $sp

        /*
         * Store LO and HI registers to preserve state of multiply/divide
         * operations.
         */
        mflo	$v0
        mfhi	$v1
        sw     $v0, OFFSET_LO($sp)
        sw     $v1, OFFSET_HI($sp)

        /*
         * Now move the previously stored EPC, Cause, BadVAddr and Status
         * to the normal stack frame.
         */
        lui    $v0, (SAFE_PLACE) >> 16
        or     $v0, (SAFE_PLACE) & 0xffff

        lw     $v1, OFFSET_SAFE_EPC($v0)
        sw     $v1, OFFSET_EPC($sp)
        lw     $v1, OFFSET_SAFE_CAUSE($v0)
        sw     $v1, OFFSET_BADVADDR($sp)
        lw     $v1, OFFSET_SAFE_BADVADDR($v0)
        sw     $v1, OFFSET_BADVADDR($sp)
        lw     $v1, OFFSET_SAFE_STATUS($v0)
        sw     $v1, OFFSET_STATUS($sp)


        /*
         * Call the C exception handler and pass the pointer
         * to the stack frame to it...
         */
        jal    exception
        or     $a0, $0, $sp


        /*
         * Move EPC and Status back to the unmapped memory.
         */
        lui    $v0, (SAFE_PLACE) >> 16
        or     $v0, (SAFE_PLACE) & 0xffff

        lw     $v1, OFFSET_STATUS($sp)
        sw     $v1, OFFSET_SAFE_STATUS($v0)
        lw     $v1, OFFSET_EPC($sp)
        sw     $v1, OFFSET_SAFE_EPC($v0)

        /*
         * Restore the LO and HI registers.
         */
        lw     $v1, OFFSET_HI($sp)
        lw     $v0, OFFSET_LO($sp)
        mthi	$v1
        mtlo	$v0

        /*
         * Restore the general registers.
         */
        RESTORE_GENERAL_REGS $sp

        /*
         * Finally restore the Status and EPC registers from the unmapped
         * memory. Restoring the Status register will switch the processor
         * back into exception level.
         */
        lui     $k1, (SAFE_PLACE) >> 16
        or      $k1, (SAFE_PLACE) & 0xffff

        lw      $k0, OFFSET_SAFE_EPC($k1)
        mtc0    $k0, $epc
        lw      $k0, OFFSET_SAFE_STATUS($k1)
        mtc0    $k0, $status

        /*
         * Release the stack frame and return from exception.
         */
        addiu   $sp, EX_STACK_FRAME
        eret

        .end    handle_general_exception



/*
 *
 * TLB Refill exception handler
 */
        .ent handle_tlbrefill

handle_tlbrefill:

        /*
         * To avoid writing to user stack, use a temporary kernel
         * stack pointed to by the k0 scratch register.
         */
        lui     $k0, (TEMP_STACK_TOP + TEMP_STACK_SIZE) >> 16
        or      $k0, (TEMP_STACK_TOP + TEMP_STACK_SIZE) & 0xffff


        /*
         * Allocate a frame on the temporary stack and store the user
         * stack pointer and all general registers on the stack. After that
         * we can use the stack pointer. Registers k0 and k1 are supposed to
         * be stored and restored.
         */
        addiu   $k0, -EX_STACK_FRAME

        sw      $sp, OFFSET_SP($k0)
        STORE_GENERAL_REGS $k0
        or      $sp, $0, $k0

        /*
         * Store LO and HI registers to preserve state of multiply/divide
         * operations.
         */
        mflo    $v0
        mfhi    $v1
        sw      $v0, OFFSET_LO($sp)
        sw      $v1, OFFSET_HI($sp)

        /*
         * Call the C exception handler...
         */
        jal tlbrefill
        nop

        /*
         * Restore the LO and HI registers.
         */
        lw      $v1, OFFSET_HI($sp)
        lw      $v0, OFFSET_LO($sp)
        mthi    $v1
        mtlo    $v0

        /*
         * Restore the general registers and the user stack pointer.
         */
        RESTORE_GENERAL_REGS $sp
        lw	$sp, OFFSET_SP($sp)

        /*
         * Return from exception.
         */
        eret

        .end    handle_tlbrefill



/*
 * switch_cpu_context
 *
 * Switches processor context from one thread to another. The first argument
 * points to a pointer to the top of the stack of the current thread, the
 * second argument points to a kernel stack of the target thread.
 *
 * Note: The processor context switch does not include memory management context.
 * Note: Interrupts must be disabled during the call.
 */
        .globl  switch_cpu_context
        .ent    switch_cpu_context

switch_cpu_context:
        /*
         * Allocate a frame on the stack and update the address of the
         * stack top of the current thread.
         */
        addiu   $sp, -EX_STACK_FRAME
        sw      $sp, ($a0)

        /*
         * Store general registers on the stack.
         */
        STORE_GENERAL_REGS $sp

        /*
         * Store HI, LO, and Status registers.
         */
        mflo    $v0
        mfhi    $v1
        sw      $v0, OFFSET_LO($sp)
        sw      $v1, OFFSET_HI($sp)
        mfc0    $t0, $status
        sw      $t0, OFFSET_STATUS($sp)

        /*
         * Switch to the stak of the new thread.
         */
        or      $sp, $0, $a1

        /*
         * Load the new HI and LO registers.
         */
         lw     $v1, OFFSET_HI($sp)
         lw     $v0, OFFSET_LO($sp)
         mthi	$v1
         mtlo	$v0

        /*
         * Restore the status register.
         */
        lw      $a0, OFFSET_STATUS($sp)
        mtc0    $a0, $status
        
        /*
         * Load the new content of general registers.
         */
        RESTORE_GENERAL_REGS $sp

        
        /*
         * Continue in what the target thread was doing
         * when it was suspended.
         */
        j       $ra
        addiu   $sp, EX_STACK_FRAME

        .end    switch_cpu_context

